@@ -1,7 +1,7 @@
 #![forbid(unsafe_code)]
 
-extern crate base64;
 extern crate charset;
+extern crate data_encoding;
 extern crate quoted_printable;
 
 use std::borrow::Cow;
@@ -9,7 +9,7 @@ use std::collections::{BTreeMap, HashMap};
 use std::error;
 use std::fmt;
 
-use charset::{Charset, decode_latin1};
+use charset::{decode_latin1, Charset};
 
 mod addrparse;
 pub mod body;
@@ -36,7 +36,7 @@ pub enum MailParseError {
     QuotedPrintableDecodeError(quoted_printable::QuotedPrintableError),
     /// Data that was specified as being in the base64 transfer-encoding could
     /// not be successfully decoded as base64 data.
-    Base64DecodeError(base64::DecodeError),
+    Base64DecodeError(data_encoding::DecodeError),
     /// An error occurred when converting the raw byte data to Rust UTF-8 string
     /// format using the charset specified in the message.
     EncodingError(std::borrow::Cow<'static, str>),
@@ -82,8 +82,8 @@ impl From<quoted_printable::QuotedPrintableError> for MailParseError {
     }
 }
 
-impl From<base64::DecodeError> for MailParseError {
-    fn from(err: base64::DecodeError) -> MailParseError {
+impl From<data_encoding::DecodeError> for MailParseError {
+    fn from(err: data_encoding::DecodeError) -> MailParseError {
         MailParseError::Base64DecodeError(err)
     }
 }
@@ -154,6 +154,30 @@ fn test_find_from_u8() {
     assert_eq!(find_from_u8(b"hello world", 10, b"d"), None);
 }
 
+// Like find_from_u8, but additionally filters such that `key` is at the start
+// of a line (preceded by `\n`) or at the start of the search space.
+fn find_from_u8_line_prefix(line: &[u8], ix_start: usize, key: &[u8]) -> Option<usize> {
+    let mut start = ix_start;
+    while let Some(ix) = find_from_u8(line, start, key) {
+        if ix == ix_start || line[ix - 1] == b'\n' {
+            return Some(ix);
+        }
+        start = ix + 1;
+    }
+    None
+}
+
+#[test]
+fn test_find_from_u8_line_prefix() {
+    assert_eq!(find_from_u8_line_prefix(b"hello world", 0, b"he"), Some(0));
+    assert_eq!(find_from_u8_line_prefix(b"hello\nhello", 0, b"he"), Some(0));
+    assert_eq!(find_from_u8_line_prefix(b"hello\nhello", 1, b"he"), Some(6));
+    assert_eq!(find_from_u8_line_prefix(b"hello world", 0, b"wo"), None);
+    assert_eq!(find_from_u8_line_prefix(b"hello\nworld", 0, b"wo"), Some(6));
+    assert_eq!(find_from_u8_line_prefix(b"hello\nworld", 6, b"wo"), Some(6));
+    assert_eq!(find_from_u8_line_prefix(b"hello\nworld", 7, b"wo"), None);
+}
+
 impl<'a> MailHeader<'a> {
     /// Get the name of the header. Note that header names are case-insensitive.
     /// Prefer using get_key_ref where possible for better performance.
@@ -167,11 +191,27 @@ impl<'a> MailHeader<'a> {
         decode_latin1(self.key)
     }
 
+    pub(crate) fn decode_utf8_or_latin1(&'a self) -> Cow<'a, str> {
+        // RFC 6532 says that header values can be UTF-8. Let's try that first, and
+        // fall back to latin1 if that fails, for better backwards-compatibility with
+        // older versions of this library that didn't try UTF-8.
+        match std::str::from_utf8(self.value) {
+            Ok(s) => Cow::Borrowed(s),
+            Err(_) => decode_latin1(self.value),
+        }
+    }
+
     /// Get the value of the header. Any sequences of newlines characters followed
     /// by whitespace are collapsed into a single space. In effect, header values
     /// wrapped across multiple lines are compacted back into one line, while
     /// discarding the extra whitespace required by the MIME format. Additionally,
     /// any quoted-printable words in the value are decoded.
+    /// Note that this function attempts to decode the header value bytes as UTF-8
+    /// first, and falls back to Latin-1 if the UTF-8 decoding fails. This attempts
+    /// to be compliant with both RFC 6532 as well as older versions of this library.
+    /// To avoid the Latin-1 fallback decoding, which may end up returning "garbage",
+    /// prefer using the get_value_utf8 function instead, which will fail and return
+    /// an error instead of falling back to Latin-1.
     ///
     /// # Examples
     /// ```
@@ -181,9 +221,13 @@ impl<'a> MailHeader<'a> {
     ///     assert_eq!(parsed.get_value(), "\u{a1}Hola, se\u{f1}or!");
     /// ```
     pub fn get_value(&self) -> String {
+        let chars = self.decode_utf8_or_latin1();
+        self.normalize_header(chars)
+    }
+
+    fn normalize_header(&'a self, chars: Cow<'a, str>) -> String {
         let mut result = String::new();
 
-        let chars = decode_latin1(self.value);
         for tok in header::normalized_tokens(&chars) {
             match tok {
                 HeaderToken::Text(t) => {
@@ -205,6 +249,29 @@ impl<'a> MailHeader<'a> {
         result
     }
 
+    /// Get the value of the header. Any sequences of newlines characters followed
+    /// by whitespace are collapsed into a single space. In effect, header values
+    /// wrapped across multiple lines are compacted back into one line, while
+    /// discarding the extra whitespace required by the MIME format. Additionally,
+    /// any quoted-printable words in the value are decoded. As per RFC 6532, this
+    /// function assumes the raw header value is encoded as UTF-8, and does that
+    /// decoding prior to tokenization and other processing. An EncodingError is
+    /// returned if the raw header value cannot be decoded as UTF-8.
+    ///
+    /// # Examples
+    /// ```
+    ///     use mailparse::parse_header;
+    ///     let (parsed, _) = parse_header(b"Subject: \xC2\xA1Hola, se\xC3\xB1or!").unwrap();
+    ///     assert_eq!(parsed.get_key(), "Subject");
+    ///     assert_eq!(parsed.get_value(), "\u{a1}Hola, se\u{f1}or!");
+    /// ```
+    pub fn get_value_utf8(&self) -> Result<String, MailParseError> {
+        let chars = std::str::from_utf8(self.value).map_err(|_| {
+            MailParseError::EncodingError(Cow::Borrowed("Invalid UTF-8 in header value"))
+        })?;
+        Ok(self.normalize_header(Cow::Borrowed(chars)))
+    }
+
     /// Get the raw, unparsed value of the header key.
     ///
     /// # Examples
@@ -520,6 +587,16 @@ impl Default for ParsedContentType {
     }
 }
 
+impl ParsedContentType {
+    fn default_conditional(in_multipart_digest: bool) -> Self {
+        let mut default = Self::default();
+        if in_multipart_digest {
+            default.mimetype = "message/rfc822".to_string();
+        }
+        default
+    }
+}
+
 /// Helper method to parse a header value as a Content-Type header. Note that
 /// the returned object's `params` map will contain a charset key if a charset
 /// was explicitly specified in the header; otherwise the `params` map will not
@@ -651,6 +728,8 @@ pub fn parse_content_disposition(header: &str) -> ParsedContentDisposition {
 /// a vector of other ParsedMail structures for the subparts.
 #[derive(Debug)]
 pub struct ParsedMail<'a> {
+    /// The raw bytes that make up this message (or subpart).
+    pub raw_bytes: &'a [u8],
     /// The raw bytes that make up the header block for this message (or subpart).
     header_bytes: &'a [u8],
     /// The headers for the message (or message subpart).
@@ -785,6 +864,37 @@ impl<'a> ParsedMail<'a> {
             .unwrap_or_default();
         disposition
     }
+
+    /// Returns a depth-first pre-order traversal of the subparts of
+    /// this ParsedMail instance. The first item returned will be this
+    /// ParsedMail itself.
+    pub fn parts(&'a self) -> PartsIterator<'a> {
+        PartsIterator {
+            parts: vec![&self],
+            index: 0,
+        }
+    }
+}
+
+pub struct PartsIterator<'a> {
+    parts: Vec<&'a ParsedMail<'a>>,
+    index: usize,
+}
+
+impl<'a> Iterator for PartsIterator<'a> {
+    type Item = &'a ParsedMail<'a>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        if self.index >= self.parts.len() {
+            return None;
+        }
+
+        let cur = self.parts[self.index];
+        self.index += 1;
+        self.parts
+            .splice(self.index..self.index, cur.subparts.iter());
+        Some(cur)
+    }
 }
 
 /// The main mail-parsing entry point.
@@ -825,13 +935,21 @@ impl<'a> ParsedMail<'a> {
 ///     assert_eq!(dateparse(parsed.headers.get_first_value("Date").unwrap().as_str()).unwrap(), 1475417182);
 /// ```
 pub fn parse_mail(raw_data: &[u8]) -> Result<ParsedMail, MailParseError> {
+    parse_mail_recursive(raw_data, false)
+}
+
+fn parse_mail_recursive(
+    raw_data: &[u8],
+    in_multipart_digest: bool,
+) -> Result<ParsedMail, MailParseError> {
     let (headers, ix_body) = parse_headers(raw_data)?;
     let ctype = headers
         .get_first_value("Content-Type")
         .map(|s| parse_content_type(&s))
-        .unwrap_or_default();
+        .unwrap_or_else(|| ParsedContentType::default_conditional(in_multipart_digest));
 
     let mut result = ParsedMail {
+        raw_bytes: raw_data,
         header_bytes: &raw_data[0..ix_body],
         headers,
         ctype,
@@ -842,20 +960,24 @@ pub fn parse_mail(raw_data: &[u8]) -> Result<ParsedMail, MailParseError> {
         && result.ctype.params.get("boundary").is_some()
         && raw_data.len() > ix_body
     {
+        let in_multipart_digest = result.ctype.mimetype == "multipart/digest";
         let boundary = String::from("--") + &result.ctype.params["boundary"];
-        if let Some(ix_body_end) = find_from_u8(raw_data, ix_body, boundary.as_bytes()) {
+        if let Some(ix_body_end) = find_from_u8_line_prefix(raw_data, ix_body, boundary.as_bytes())
+        {
             result.body_bytes = &raw_data[ix_body..ix_body_end];
             let mut ix_boundary_end = ix_body_end + boundary.len();
             while let Some(ix_part_start) =
                 find_from_u8(raw_data, ix_boundary_end, b"\n").map(|v| v + 1)
             {
                 // if there is no terminating boundary, assume the part end is the end of the email
-                let ix_part_end = find_from_u8(raw_data, ix_part_start, boundary.as_bytes())
-                    .unwrap_or_else(|| raw_data.len());
+                let ix_part_end =
+                    find_from_u8_line_prefix(raw_data, ix_part_start, boundary.as_bytes())
+                        .unwrap_or_else(|| raw_data.len());
 
-                result
-                    .subparts
-                    .push(parse_mail(&raw_data[ix_part_start..ix_part_end])?);
+                result.subparts.push(parse_mail_recursive(
+                    &raw_data[ix_part_start..ix_part_end],
+                    in_multipart_digest,
+                )?);
                 ix_boundary_end = ix_part_end + boundary.len();
                 if ix_boundary_end + 2 > raw_data.len()
                     || (raw_data[ix_boundary_end] == b'-' && raw_data[ix_boundary_end + 1] == b'-')
@@ -958,7 +1080,10 @@ fn parse_param_content(content: &str) -> ParamContent {
 /// containing the encoding (or empty string for no encoding found) and a flag
 /// that indicates if the encoding needs to be stripped from the value. This is
 /// set to true for non-continuation parameter values.
-fn compute_parameter_encodings(map: &BTreeMap<String, String>, decode_key_list: &Vec<String>) -> HashMap<String, (String, bool)> {
+fn compute_parameter_encodings(
+    map: &BTreeMap<String, String>,
+    decode_key_list: &Vec<String>,
+) -> HashMap<String, (String, bool)> {
     // To handle section 4.1 (combining encodings with continuations), we first
     // compute the encoding for each parameter value or parameter value segment
     // that is encoded. For continuation segments the encoding from the *0 segment
@@ -968,18 +1093,32 @@ fn compute_parameter_encodings(map: &BTreeMap<String, String>, decode_key_list:
         if let Some(unwrap_key) = decode_key.strip_suffix("*0") {
             // Per spec, there should always be an encoding. If it's missing, handle that case gracefully
             // by setting it to an empty string that we handle specially later.
-            let encoding = map.get(&format!("{}*", decode_key)).unwrap().split('\'').next().unwrap_or("");
+            let encoding = map
+                .get(&format!("{}*", decode_key))
+                .unwrap()
+                .split('\'')
+                .next()
+                .unwrap_or("");
             let continuation_prefix = format!("{}*", unwrap_key);
             for continuation_key in decode_key_list {
                 if continuation_key.starts_with(&continuation_prefix) {
                     // This may (intentionally) overwite encodings previously found for the
                     // continuation segments (which are bogus). In those cases, the flag
                     // in the tuple should get updated from true to false.
-                    encodings.insert(continuation_key.clone(), (encoding.to_string(), continuation_key == decode_key));
+                    encodings.insert(
+                        continuation_key.clone(),
+                        (encoding.to_string(), continuation_key == decode_key),
+                    );
                 }
             }
         } else if !encodings.contains_key(decode_key) {
-            let encoding = map.get(&format!("{}*", decode_key)).unwrap().split('\'').next().unwrap_or("").to_string();
+            let encoding = map
+                .get(&format!("{}*", decode_key))
+                .unwrap()
+                .split('\'')
+                .next()
+                .unwrap_or("")
+                .to_string();
             let old_value = encodings.insert(decode_key.clone(), (encoding, true));
             assert!(old_value.is_none());
         }
@@ -1565,13 +1704,16 @@ mod tests {
 
     #[test]
     fn test_parameter_value_continuations() {
-        let parsed = parse_param_content("attachment;\n\tfilename*0=\"X\";\n\tfilename*1=\"Y.pdf\"");
+        let parsed =
+            parse_param_content("attachment;\n\tfilename*0=\"X\";\n\tfilename*1=\"Y.pdf\"");
         assert_eq!(parsed.value, "attachment");
         assert_eq!(parsed.params["filename"], "XY.pdf");
         assert_eq!(parsed.params.contains_key("filename*0"), false);
         assert_eq!(parsed.params.contains_key("filename*1"), false);
 
-        let parsed = parse_param_content("attachment;\n\tfilename=XX.pdf;\n\tfilename*0=\"X\";\n\tfilename*1=\"Y.pdf\"");
+        let parsed = parse_param_content(
+            "attachment;\n\tfilename=XX.pdf;\n\tfilename*0=\"X\";\n\tfilename*1=\"Y.pdf\"",
+        );
         assert_eq!(parsed.value, "attachment");
         assert_eq!(parsed.params["filename"], "XX.pdf");
         assert_eq!(parsed.params["filename*0"], "X");
@@ -1587,7 +1729,10 @@ mod tests {
         let parsed = parse_param_content("attachment;\n\tfilename*0*=us-ascii''%28X%29%20801%20-%20X;\n\tfilename*1*=%20%E2%80%93%20X%20;\n\tfilename*2*=X%20X%2Epdf");
         // Note this is a real-world case from mutt, but it's wrong. The original filename had an en dash \u{2013} but mutt
         // declared us-ascii as the encoding instead of utf-8 for some reason.
-        assert_eq!(parsed.params["filename"], "(X) 801 - X \u{00E2}\u{20AC}\u{201C} X X X.pdf");
+        assert_eq!(
+            parsed.params["filename"],
+            "(X) 801 - X \u{00E2}\u{20AC}\u{201C} X X X.pdf"
+        );
         assert_eq!(parsed.params.contains_key("filename*0*"), false);
         assert_eq!(parsed.params.contains_key("filename*0"), false);
         assert_eq!(parsed.params.contains_key("filename*1*"), false);
@@ -1616,32 +1761,36 @@ mod tests {
         assert_eq!(parsed.params["filename*"], "nonexistent'foo'%e2%80%a1.bin");
         assert_eq!(parsed.params.contains_key("filename"), false);
 
-        let parsed = parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*1*=%e2%80%A1.bin");
+        let parsed = parse_param_content(
+            "attachment; filename*0*=utf-8'en'%e2%80%a1; filename*1*=%e2%80%A1.bin",
+        );
         assert_eq!(parsed.params["filename"], "\u{2021}\u{2021}.bin");
         assert_eq!(parsed.params.contains_key("filename*0*"), false);
         assert_eq!(parsed.params.contains_key("filename*0"), false);
         assert_eq!(parsed.params.contains_key("filename*1*"), false);
         assert_eq!(parsed.params.contains_key("filename*1"), false);
 
-        let parsed = parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*1=%20.bin");
+        let parsed =
+            parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*1=%20.bin");
         assert_eq!(parsed.params["filename"], "\u{2021}%20.bin");
         assert_eq!(parsed.params.contains_key("filename*0*"), false);
         assert_eq!(parsed.params.contains_key("filename*0"), false);
         assert_eq!(parsed.params.contains_key("filename*1*"), false);
         assert_eq!(parsed.params.contains_key("filename*1"), false);
 
-        let parsed = parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*2*=%20.bin");
+        let parsed =
+            parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*2*=%20.bin");
         assert_eq!(parsed.params["filename"], "\u{2021}");
         assert_eq!(parsed.params["filename*2"], " .bin");
         assert_eq!(parsed.params.contains_key("filename*0*"), false);
         assert_eq!(parsed.params.contains_key("filename*0"), false);
         assert_eq!(parsed.params.contains_key("filename*2*"), false);
 
-        let parsed = parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*0=foo.bin");
+        let parsed =
+            parse_param_content("attachment; filename*0*=utf-8'en'%e2%80%a1; filename*0=foo.bin");
         assert_eq!(parsed.params["filename"], "foo.bin");
         assert_eq!(parsed.params["filename*0*"], "utf-8'en'%e2%80%a1");
         assert_eq!(parsed.params.contains_key("filename*0"), false);
-
     }
 
     #[test]
@@ -1712,6 +1861,22 @@ mod tests {
         };
     }
 
+    #[test]
+    fn test_base64_content_encoding_multiple_strings() {
+        let mail = parse_mail(
+            b"Content-Transfer-Encoding: base64\r\n\r\naGVsbG 8gd\r\n29ybGQ=\r\nZm9vCg==",
+        )
+        .unwrap();
+        match mail.get_body_encoded() {
+            Body::Base64(body) => {
+                assert_eq!(body.get_raw(), b"aGVsbG 8gd\r\n29ybGQ=\r\nZm9vCg==");
+                assert_eq!(body.get_decoded().unwrap(), b"hello worldfoo\n");
+                assert_eq!(body.get_decoded_as_string().unwrap(), "hello worldfoo\n");
+            }
+            _ => assert!(false),
+        };
+    }
+
     #[test]
     fn test_binary_content_encoding() {
         let mail = parse_mail(b"Content-Transfer-Encoding: binary\r\n\r\n######").unwrap();
@@ -1769,7 +1934,7 @@ mod tests {
     fn test_fuzzer_testcase() {
         const INPUT: &'static str = "U3ViamVjdDplcy1UeXBlOiBtdW50ZW50LVV5cGU6IW11bAAAAAAAAAAAamVjdDplcy1UeXBlOiBtdW50ZW50LVV5cGU6IG11bAAAAAAAAAAAAAAAAABTTUFZdWJqZf86OiP/dCBTdWJqZWN0Ol8KRGF0ZTog/////////////////////wAAAAAAAAAAAHQgYnJmAHQgYnJmZXItRW5jeXBlOnY9NmU3OjA2OgAAAAAAAAAAAAAAADEAAAAAAP/8mAAAAAAAAAAA+f///wAAAAAAAP8AAAAAAAAAAAAAAAAAAAAAAAAAPT0/PzEAAAEAAA==";
 
-        if let Ok(parsed) = parse_mail(&base64::decode(INPUT).unwrap()) {
+        if let Ok(parsed) = parse_mail(&data_encoding::BASE64.decode(INPUT.as_bytes()).unwrap()) {
             if let Some(date) = parsed.headers.get_first_value("Date") {
                 let _ = dateparse(&date);
             }
@@ -1779,7 +1944,7 @@ mod tests {
     #[test]
     fn test_fuzzer_testcase_2() {
         const INPUT: &'static str = "U3ViamVjdDogVGhpcyBpcyBhIHRlc3QgZW1haWwKQ29udGVudC1UeXBlOiBtdWx0aXBhcnQvYWx0ZXJuYXRpdmU7IGJvdW5kYXJ5PczMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMzMZm9vYmFyCkRhdGU6IFN1biwgMDIgT2MKCi1TdWJqZWMtZm9vYmFydDo=";
-        if let Ok(parsed) = parse_mail(&base64::decode(INPUT).unwrap()) {
+        if let Ok(parsed) = parse_mail(&data_encoding::BASE64.decode(INPUT.as_bytes()).unwrap()) {
             if let Some(date) = parsed.headers.get_first_value("Date") {
                 let _ = dateparse(&date);
             }
@@ -1799,4 +1964,186 @@ mod tests {
     fn test_percent_decoder() {
         assert_eq!(percent_decode("hi %0d%0A%%2A%zz%"), b"hi \r\n%*%zz%");
     }
+
+    #[test]
+    fn test_default_content_type_in_multipart_digest() {
+        // Per https://datatracker.ietf.org/doc/html/rfc2046#section-5.1.5
+        let mail = parse_mail(
+            concat!(
+                "Content-Type: multipart/digest; boundary=myboundary\r\n\r\n",
+                "--myboundary\r\n\r\n",
+                "blah blah blah\r\n\r\n",
+                "--myboundary--\r\n"
+            )
+            .as_bytes(),
+        )
+        .unwrap();
+        assert_eq!(mail.headers.len(), 1);
+        assert_eq!(mail.ctype.mimetype, "multipart/digest");
+        assert_eq!(mail.subparts[0].headers.len(), 0);
+        assert_eq!(mail.subparts[0].ctype.mimetype, "message/rfc822");
+
+        let mail = parse_mail(
+            concat!(
+                "Content-Type: multipart/whatever; boundary=myboundary\n",
+                "\n",
+                "--myboundary\n",
+                "\n",
+                "blah blah blah\n",
+                "--myboundary\n",
+                "Content-Type: multipart/digest; boundary=nestedboundary\n",
+                "\n",
+                "--nestedboundary\n",
+                "\n",
+                "nested default part\n",
+                "--nestedboundary\n",
+                "Content-Type: text/html\n",
+                "\n",
+                "nested html part\n",
+                "--nestedboundary\n",
+                "Content-Type: multipart/insidedigest; boundary=insideboundary\n",
+                "\n",
+                "--insideboundary\n",
+                "\n",
+                "inside part\n",
+                "--insideboundary--\n",
+                "--nestedboundary--\n",
+                "--myboundary--\n"
+            )
+            .as_bytes(),
+        )
+        .unwrap();
+        let mut parts = mail.parts();
+        let mut part = parts.next().unwrap(); // mail
+
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "multipart/whatever");
+
+        part = parts.next().unwrap(); // mail.subparts[0]
+        assert_eq!(part.headers.len(), 0);
+        assert_eq!(part.ctype.mimetype, "text/plain");
+        assert_eq!(part.get_body_raw().unwrap(), b"blah blah blah\n");
+
+        part = parts.next().unwrap(); // mail.subparts[1]
+        assert_eq!(part.ctype.mimetype, "multipart/digest");
+
+        part = parts.next().unwrap(); // mail.subparts[1].subparts[0]
+        assert_eq!(part.headers.len(), 0);
+        assert_eq!(part.ctype.mimetype, "message/rfc822");
+        assert_eq!(part.get_body_raw().unwrap(), b"nested default part\n");
+
+        part = parts.next().unwrap(); // mail.subparts[1].subparts[1]
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "text/html");
+        assert_eq!(part.get_body_raw().unwrap(), b"nested html part\n");
+
+        part = parts.next().unwrap(); // mail.subparts[1].subparts[2]
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "multipart/insidedigest");
+
+        part = parts.next().unwrap(); // mail.subparts[1].subparts[2].subparts[0]
+        assert_eq!(part.headers.len(), 0);
+        assert_eq!(part.ctype.mimetype, "text/plain");
+        assert_eq!(part.get_body_raw().unwrap(), b"inside part\n");
+
+        assert!(parts.next().is_none());
+    }
+
+    #[test]
+    fn boundary_is_suffix_of_another_boundary() {
+        // From https://github.com/staktrace/mailparse/issues/100
+        let mail = parse_mail(
+            concat!(
+                "Content-Type: multipart/mixed; boundary=\"section_boundary\"\n",
+                "\n",
+                "--section_boundary\n",
+                "Content-Type: multipart/alternative; boundary=\"--section_boundary\"\n",
+                "\n",
+                "----section_boundary\n",
+                "Content-Type: text/html;\n",
+                "\n",
+                "<em>Good evening!</em>\n",
+                "----section_boundary\n",
+                "Content-Type: text/plain;\n",
+                "\n",
+                "Good evening!\n",
+                "----section_boundary\n",
+                "--section_boundary\n"
+            )
+            .as_bytes(),
+        )
+        .unwrap();
+
+        let mut parts = mail.parts();
+        let mut part = parts.next().unwrap(); // mail
+
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "multipart/mixed");
+        assert_eq!(part.subparts.len(), 1);
+
+        part = parts.next().unwrap(); // mail.subparts[0]
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "multipart/alternative");
+        assert_eq!(part.subparts.len(), 2);
+
+        part = parts.next().unwrap(); // mail.subparts[0].subparts[0]
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "text/html");
+        assert_eq!(part.get_body_raw().unwrap(), b"<em>Good evening!</em>\n");
+        assert_eq!(part.subparts.len(), 0);
+
+        part = parts.next().unwrap(); // mail.subparts[0].subparts[1]
+        assert_eq!(part.headers.len(), 1);
+        assert_eq!(part.ctype.mimetype, "text/plain");
+        assert_eq!(part.get_body_raw().unwrap(), b"Good evening!\n");
+        assert_eq!(part.subparts.len(), 0);
+
+        assert!(parts.next().is_none());
+    }
+
+    #[test]
+    fn test_parts_iterator() {
+        let mail = parse_mail(
+            concat!(
+                "Content-Type: multipart/mixed; boundary=\"top_boundary\"\n",
+                "\n",
+                "--top_boundary\n",
+                "Content-Type: multipart/alternative; boundary=\"internal_boundary\"\n",
+                "\n",
+                "--internal_boundary\n",
+                "Content-Type: text/html;\n",
+                "\n",
+                "<em>Good evening!</em>\n",
+                "--internal_boundary\n",
+                "Content-Type: text/plain;\n",
+                "\n",
+                "Good evening!\n",
+                "--internal_boundary\n",
+                "--top_boundary\n",
+                "Content-Type: text/unknown;\n",
+                "\n",
+                "You read this?\n",
+                "--top_boundary\n"
+            )
+            .as_bytes(),
+        )
+        .unwrap();
+
+        let mut parts = mail.parts();
+        assert_eq!(parts.next().unwrap().ctype.mimetype, "multipart/mixed");
+        assert_eq!(
+            parts.next().unwrap().ctype.mimetype,
+            "multipart/alternative"
+        );
+        assert_eq!(parts.next().unwrap().ctype.mimetype, "text/html");
+        assert_eq!(parts.next().unwrap().ctype.mimetype, "text/plain");
+        assert_eq!(parts.next().unwrap().ctype.mimetype, "text/unknown");
+        assert!(parts.next().is_none());
+
+        let mail = parse_mail(concat!("Content-Type: text/plain\n").as_bytes()).unwrap();
+
+        let mut parts = mail.parts();
+        assert_eq!(parts.next().unwrap().ctype.mimetype, "text/plain");
+        assert!(parts.next().is_none());
+    }
 }
